/* Copyright (c) 1993 xTech */

/* AC: Beware, this is exptremely tricky code. Do  not  modify  it!
 * I put special efforts to chart optimizing compilers. This module
 * should not be optimized, and even the order of procedures is  of
 * importance.
 *
 * The problems are:
 *   1. We do not know which direction the stack is growing up.
 *   2. We do not know where the hardware registers are in 'jmp_buf'.
 *   3. There could be a problem on segmented architectures when
 *      stack and data segments are separated, and there  is  no
 *      way to access these segments directly/indirectly from C.
 *
 * This module detect all hardware parameters (to be exact, 1 and 2)
 * on the fly. For case 3) there is necessary to use additional code
 * for any particular case (see Coroutines.NEWPROCESS).
 *
 * I supposed that PC (program counter) is incrementing to fetch the
 * next instruction.
 *
 * Please contact RTA if you have any problems.
 *
 */

#include <stdio.h>
#include <setjmp.h>
#include "Coroutines.h"

#define MASK 0xffffff /* mask for PC */
#define SIZE (sizeof (jmp_buf) / sizeof (int))
#define DIFF (128*sizeof(int))
#define MAGIC 0x67CA

typedef struct
{
  int magic;
  jmp_buf buf;
}
coroutine;

/* Must be global */
union
{
  jmp_buf j;
  unsigned u[SIZE];
}
Coroutines_1_jmpbuf;

static char kind[SIZE];

typedef unsigned (X2C_PROCLASS * _setup)(unsigned pc);
_setup Coroutines_1_setup[3];
char * (X2C_PROCLASS * Coroutines_1_direc)(unsigned cnt);
X2C_PROC Coroutines_1_empty;

/* local working variables */
static unsigned loc_pc;
static unsigned loc_sp;
static unsigned loc_cs;
static unsigned loc_ss;
static int loc_dir;
static int loc_endian;
static coroutine *loc_current;

static unsigned X2C_PROCLASS diff (unsigned i, unsigned j)
{
  if (i > j) return (i-j);
  return (j-i);
}

unsigned X2C_PROCLASS Coroutines_1__setup (unsigned pc)
{
  unsigned n;
  n = (unsigned)(&n);
  if (diff (pc, n) > DIFF*2 && n != loc_cs && n != loc_ss)
  {
    if (!setjmp (Coroutines_1_jmpbuf.j)) (*Coroutines_1_empty)();
  }
  else n = (*Coroutines_1_setup[0])(pc);
  if (n == 0) ++n;
  return (n);
}

char * X2C_PROCLASS Coroutines_1__direc (unsigned n)
{
  char *p;
  p = (char *)(&p);
  if (n > 0) p = (*Coroutines_1_direc)(n-1);
  return (p);
}

void X2C_PROCLASS Coroutines_1__empty (void)
{
}

static unsigned X2C_PROCLASS highproc (X2C_PROC p)
{
  if (sizeof (p) <= sizeof (unsigned)) return (0);
  return ((unsigned)(((unsigned long)p)>>(sizeof(unsigned)*8)));
}

static unsigned X2C_PROCLASS highaddr (X2C_ADDRESS p)
{
  if (sizeof (p) <= sizeof (unsigned)) return (0);
  return ((unsigned)(((unsigned long)p)>>(sizeof(unsigned)*8)));
}

static void X2C_PROCLASS fatal (char *s)
{
  fprintf (stderr, "Coroutines fatal error: %s\n", s);
  X2C_HALT (X2C_FATAL_ERROR);
}

void X2C_PROCLASS Coroutines_1_fatal (void)
{
  fatal ("coroutine ended without transfer");
}

static void X2C_PROCLASS nonexist (void)
{
  fatal ("transfer to non-existing process");
}

static coroutine * X2C_PROCLASS allocate (Coroutines_Coroutine *addr)
{
  coroutine *p;
  if ((p = (coroutine *)malloc (sizeof (*p))) == 0) fatal ("no enough memory");
  p->magic = MAGIC;
  if (setjmp (p->buf)) nonexist ();
  return (*(coroutine **)addr = p);
}

void X2C_PROCLASS Coroutines_NEWPROCESS (X2C_PROC proc, X2C_ADDRESS space, X2C_INDEX size, Coroutines_Coroutine *_this)
{
  coroutine *this = allocate (_this);
  unsigned *addr;
  unsigned pc, cs, sp, ss;
  int i;

  if (setjmp (this->buf)) nonexist ();

  size = (size+sizeof(unsigned)-1)&~((X2C_INDEX)sizeof(unsigned)-1);
  if (size < DIFF) fatal ("working area size is too small");

  cs = highproc ((X2C_PROC)Coroutines_1_fatal);
  pc = (unsigned)Coroutines_1_fatal;

  size = (unsigned)size / sizeof(unsigned);
  if (loc_dir > 0)
  {
    addr = (unsigned *)space;
    for (i = 0; i < 8; ++i) addr[i] = pc;
    if (cs != 0) for (i = loc_endian; i < 8; i += 2) addr[i] = cs;
    addr += 4;
  }
  else
  {
    addr = ((unsigned *)space) + size;
    for (i = 0; i < 8; ++i) addr[i-8] = pc;
    if (cs != 0) for (i = loc_endian; i < 8; i += 2) addr[i-8] = cs;
    addr -= 4;
  }

  cs = highproc (proc);
  pc = (unsigned)proc;
  ss = highaddr ((X2C_ADDRESS)addr);
  sp = (unsigned)addr;

  addr = (unsigned *)(this->buf);
  for (i = 0; i < SIZE; ++i)
    switch (kind[i])
    {
    case 0: if (cs != 0) addr[i] = cs; break;
    case 1: if (ss != 0) addr[i] = ss; break;
    case 2: addr[i] = pc; break;
    case 3: addr[i] = sp; break;
    }
#if (defined(__ZTC__) || defined(__SC__)) && __I86__ >= 3
  /* assign ESS = EDS; I do not know how to access them */
  addr[4] = addr[9];
#endif
}

unsigned X2C_PROCLASS Coroutines_2__setup (unsigned pc)
{
  unsigned n;
  n = (unsigned)(&n);
  if (diff (pc, n) > DIFF*2 && n != loc_cs && n != loc_ss)
  {
    if (!setjmp (Coroutines_1_jmpbuf.j)) (*Coroutines_1_empty)();
  }
  else n = (*Coroutines_1_setup[1])(pc);
  if (n == 0) ++n;
  return (n);
}

void X2C_PROCLASS Coroutines_TRANSFER (Coroutines_Coroutine *_from, Coroutines_Coroutine _to)
{
  coroutine *from, *to;
  if ((from = *(coroutine **)_from) == 0) from = allocate (_from);
  if ((to = (coroutine *)_to) == 0) nonexist ();
  if (from->magic != MAGIC || to->magic != MAGIC) fatal ("bad magic number");
  if (!setjmp (from->buf))
  {
    if (loc_current == to) setjmp (to->buf);
    else longjmp ((loc_current = to)->buf, 1);
  }
}

unsigned X2C_PROCLASS Coroutines_3__setup (unsigned pc);

/* AC: artificial intellegence starts here */
void X2C_PROCLASS Coroutines_BEGIN (void)
{
  static int init = 0;
  int i;
  unsigned *addr = Coroutines_1_jmpbuf.u;

  if (init) return; init = 1;

  loc_current = (coroutine *)0;

  Coroutines_1_setup[0] = Coroutines_1__setup;
  Coroutines_1_setup[1] = Coroutines_2__setup;
  Coroutines_1_setup[2] = Coroutines_3__setup;
  Coroutines_1_empty = Coroutines_1__empty;
  Coroutines_1_direc = Coroutines_1__direc;

  i = (int)(Coroutines_1__direc (0) - Coroutines_1__direc (1));
  loc_dir = -1; if (i < 0) loc_dir = 1;
  i = 1;
  loc_endian = *((char *)(&i));

  loc_ss = highaddr ((X2C_ADDRESS)(&i));
  i = 3;
  do
  {
    --i;
    loc_cs = highproc ((X2C_PROC)Coroutines_1_setup[i]);
    loc_pc = (unsigned)(Coroutines_1_setup[i]);
  }
  while (loc_pc == loc_cs || loc_pc == loc_ss);

  loc_sp = (*Coroutines_1_setup[i])(loc_pc);
  for (i = 0; i < SIZE; ++i)
  {
    if      (loc_cs!=0 && addr[i]==loc_cs) kind[i] = 0;
    else if (loc_ss!=0 && addr[i]==loc_ss) kind[i] = 1;
    else if (loc_dir>0 && addr[i]>=loc_sp && addr[i]-loc_sp < DIFF) kind[i] = 3;
    else if (loc_dir<0 && addr[i]<=loc_sp && loc_sp-addr[i] < DIFF) kind[i] = 3;
    else if ((addr[i]&MASK)>=(loc_pc&MASK) && ((addr[i]-loc_pc)&MASK) < DIFF) kind[i] = 2;
    else kind[i] = 4;
  }
}

Coroutines_Coroutine X2C_PROCLASS Coroutines_CURRENT (void)
{
  return ((Coroutines_Coroutine)loc_current);
}

unsigned X2C_PROCLASS Coroutines_3__setup (unsigned pc)
{
  unsigned n;
  n = (unsigned)(&n);
  if (diff (pc, n) > DIFF*2 && n != loc_cs && n != loc_ss)
  {
    if (!setjmp (Coroutines_1_jmpbuf.j)) (*Coroutines_1_empty)();
  }
  else n = (*Coroutines_1_setup[2])(pc);
  if (n == 0) ++n;
  return (n);
}
